@@ -26,9 +26,10 @@
 
 #include <sys/types.h>
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+#  include <unistd.h>
 #endif // HAVE_UNISTD_H
 #include <sys/resource.h>
+#include <sys/wait.h>
 #include <grp.h>
 
 #include <cinttypes>
@@ -38,6 +39,8 @@
 
 #include <ev.h>
 
+#include <ares.h>
+
 #include "shrpx_config.h"
 #include "shrpx_connection_handler.h"
 #include "shrpx_log_config.h"
@@ -48,10 +51,12 @@
 #include "shrpx_memcached_dispatcher.h"
 #include "shrpx_memcached_request.h"
 #include "shrpx_process.h"
-#include "shrpx_ssl.h"
+#include "shrpx_tls.h"
+#include "shrpx_log.h"
 #include "util.h"
 #include "app_helper.h"
 #include "template.h"
+#include "xsi_strerror.h"
 
 using namespace nghttp2;
 
@@ -62,35 +67,39 @@ void drop_privileges(
 #ifdef HAVE_NEVERBLEED
     neverbleed_t *nb
 #endif // HAVE_NEVERBLEED
-    ) {
-  if (getuid() == 0 && get_config()->uid != 0) {
-#ifdef HAVE_INITGROUPS
-    if (initgroups(get_config()->user.get(), get_config()->gid) != 0) {
+) {
+  std::array<char, STRERROR_BUFSIZE> errbuf;
+  auto config = get_config();
+
+  if (getuid() == 0 && config->uid != 0) {
+#ifdef HAVE_NEVERBLEED
+    if (nb) {
+      neverbleed_setuidgid(nb, config->user.c_str(), 1);
+    }
+#endif // HAVE_NEVERBLEED
+
+    if (initgroups(config->user.c_str(), config->gid) != 0) {
       auto error = errno;
       LOG(FATAL) << "Could not change supplementary groups: "
-                 << strerror(error);
+                 << xsi_strerror(error, errbuf.data(), errbuf.size());
       exit(EXIT_FAILURE);
     }
-#endif // HAVE_INITGROUPS
-    if (setgid(get_config()->gid) != 0) {
+    if (setgid(config->gid) != 0) {
       auto error = errno;
-      LOG(FATAL) << "Could not change gid: " << strerror(error);
+      LOG(FATAL) << "Could not change gid: "
+                 << xsi_strerror(error, errbuf.data(), errbuf.size());
       exit(EXIT_FAILURE);
     }
-    if (setuid(get_config()->uid) != 0) {
+    if (setuid(config->uid) != 0) {
       auto error = errno;
-      LOG(FATAL) << "Could not change uid: " << strerror(error);
+      LOG(FATAL) << "Could not change uid: "
+                 << xsi_strerror(error, errbuf.data(), errbuf.size());
       exit(EXIT_FAILURE);
     }
     if (setuid(0) != -1) {
       LOG(FATAL) << "Still have root privileges?";
       exit(EXIT_FAILURE);
     }
-#ifdef HAVE_NEVERBLEED
-    if (nb) {
-      neverbleed_setuidgid(nb, get_config()->user.get(), 1);
-    }
-#endif // HAVE_NEVERBLEED
   }
 }
 } // namespace
@@ -105,18 +114,18 @@ void graceful_shutdown(ConnectionHandler *conn_handler) {
 
   conn_handler->set_graceful_shutdown(true);
 
-  conn_handler->disable_acceptor();
-
-  // After disabling accepting new connection, disptach incoming
-  // connection in backlog.
-
+  // TODO What happens for the connections not established in the
+  // kernel?
   conn_handler->accept_pending_connection();
+  conn_handler->delete_acceptor();
 
   conn_handler->graceful_shutdown_worker();
 
-  if (get_config()->num_worker == 1) {
-    if (conn_handler->get_single_worker()->get_worker_stat()->num_connections ==
-        0) {
+  auto single_worker = conn_handler->get_single_worker();
+  if (single_worker) {
+    auto worker_stat = single_worker->get_worker_stat();
+    if (worker_stat->num_connections == 0 &&
+        worker_stat->num_close_waits == 0) {
       ev_break(conn_handler->get_loop());
     }
 
@@ -129,12 +138,13 @@ namespace {
 void reopen_log(ConnectionHandler *conn_handler) {
   LOG(NOTICE) << "Reopening log files: worker process (thread main)";
 
-  (void)reopen_log_files();
-  redirect_stderr_to_errorlog();
+  auto config = get_config();
+  auto &loggingconf = config->logging;
 
-  if (get_config()->num_worker > 1) {
-    conn_handler->worker_reopen_log_files();
-  }
+  (void)reopen_log_files(loggingconf);
+  redirect_stderr_to_errorlog(loggingconf);
+
+  conn_handler->worker_reopen_log_files();
 }
 } // namespace
 
@@ -154,7 +164,7 @@ void ipc_readcb(struct ev_loop *loop, ev_io *w, int revents) {
   if (nread == 0) {
     // IPC socket closed.  Perform immediate shutdown.
     LOG(FATAL) << "IPC socket is closed.  Perform immediate shutdown.";
-    _Exit(EXIT_FAILURE);
+    nghttp2_Exit(EXIT_FAILURE);
   }
 
   for (ssize_t i = 0; i < nread; ++i) {
@@ -170,9 +180,23 @@ void ipc_readcb(struct ev_loop *loop, ev_io *w, int revents) {
 }
 } // namespace
 
+#ifdef ENABLE_HTTP3
+namespace {
+void quic_ipc_readcb(struct ev_loop *loop, ev_io *w, int revents) {
+  auto conn_handler = static_cast<ConnectionHandler *>(w->data);
+
+  if (conn_handler->quic_ipc_read() != 0) {
+    LOG(ERROR) << "Failed to read data from QUIC IPC channel";
+
+    return;
+  }
+}
+} // namespace
+#endif // ENABLE_HTTP3
+
 namespace {
 int generate_ticket_key(TicketKey &ticket_key) {
-  ticket_key.cipher = get_config()->tls_ticket_key_cipher;
+  ticket_key.cipher = get_config()->tls.ticket.cipher;
   ticket_key.hmac = EVP_sha256();
   ticket_key.hmac_keylen = EVP_MD_size(ticket_key.hmac);
 
@@ -219,7 +243,8 @@ void renew_ticket_key_cb(struct ev_loop *loop, ev_timer *w, int revents) {
 
     auto max_tickets =
         static_cast<size_t>(std::chrono::duration_cast<std::chrono::hours>(
-                                get_config()->tls_session_timeout).count());
+                                get_config()->tls.session_timeout)
+                                .count());
 
     new_keys.resize(std::min(max_tickets, old_keys.size() + 1));
     std::copy_n(std::begin(old_keys), new_keys.size() - 1,
@@ -261,15 +286,14 @@ void memcached_get_ticket_key_cb(struct ev_loop *loop, ev_timer *w,
   auto conn_handler = static_cast<ConnectionHandler *>(w->data);
   auto dispatcher = conn_handler->get_tls_ticket_key_memcached_dispatcher();
 
-  auto req = make_unique<MemcachedRequest>();
+  auto req = std::make_unique<MemcachedRequest>();
   req->key = "nghttpx:tls-ticket-key";
-  req->op = MEMCACHED_OP_GET;
-  req->cb = [conn_handler, dispatcher, w](MemcachedRequest *req,
-                                          MemcachedResult res) {
+  req->op = MemcachedOp::GET;
+  req->cb = [conn_handler, w](MemcachedRequest *req, MemcachedResult res) {
     switch (res.status_code) {
-    case MEMCACHED_ERR_NO_ERROR:
+    case MemcachedStatusCode::NO_ERROR:
       break;
-    case MEMCACHED_ERR_EXT_NETWORK_ERROR:
+    case MemcachedStatusCode::EXT_NETWORK_ERROR:
       conn_handler->on_tls_ticket_key_network_error(w);
       return;
     default:
@@ -299,14 +323,16 @@ void memcached_get_ticket_key_cb(struct ev_loop *loop, ev_timer *w,
     auto end = p + value.size();
     p += 4;
 
+    auto &ticketconf = get_config()->tls.ticket;
+
     size_t expectedlen;
     size_t enc_keylen;
     size_t hmac_keylen;
-    if (get_config()->tls_ticket_key_cipher == EVP_aes_128_cbc()) {
+    if (ticketconf.cipher == EVP_aes_128_cbc()) {
       expectedlen = 48;
       enc_keylen = 16;
       hmac_keylen = 16;
-    } else if (get_config()->tls_ticket_key_cipher == EVP_aes_256_cbc()) {
+    } else if (ticketconf.cipher == EVP_aes_256_cbc()) {
       expectedlen = 80;
       enc_keylen = 32;
       hmac_keylen = 32;
@@ -337,17 +363,17 @@ void memcached_get_ticket_key_cb(struct ev_loop *loop, ev_timer *w,
         return;
       }
       auto key = TicketKey();
-      key.cipher = get_config()->tls_ticket_key_cipher;
+      key.cipher = ticketconf.cipher;
       key.hmac = EVP_sha256();
       key.hmac_keylen = hmac_keylen;
 
-      std::copy_n(p, key.data.name.size(), key.data.name.data());
+      std::copy_n(p, key.data.name.size(), std::begin(key.data.name));
       p += key.data.name.size();
 
-      std::copy_n(p, enc_keylen, key.data.enc_key.data());
+      std::copy_n(p, enc_keylen, std::begin(key.data.enc_key));
       p += enc_keylen;
 
-      std::copy_n(p, hmac_keylen, key.data.hmac_key.data());
+      std::copy_n(p, hmac_keylen, std::begin(key.data.hmac_key));
       p += hmac_keylen;
 
       ticket_keys->keys.push_back(std::move(key));
@@ -372,73 +398,120 @@ void nb_child_cb(struct ev_loop *loop, ev_child *w, int revents) {
 
   ev_child_stop(loop, w);
 
-  LOG(FATAL) << "neverbleed process exitted; aborting now";
+  LOG(FATAL) << "neverbleed process exited; aborting now";
 
-  _Exit(EXIT_FAILURE);
+  nghttp2_Exit(EXIT_FAILURE);
 }
 } // namespace
 #endif // HAVE_NEVERBLEED
 
+namespace {
+int send_ready_event(int ready_ipc_fd) {
+  std::array<char, STRERROR_BUFSIZE> errbuf;
+  auto pid = getpid();
+  ssize_t nwrite;
+
+  while ((nwrite = write(ready_ipc_fd, &pid, sizeof(pid))) == -1 &&
+         errno == EINTR)
+    ;
+
+  if (nwrite < 0) {
+    auto error = errno;
+
+    LOG(ERROR) << "Writing PID to ready IPC channel failed: "
+               << xsi_strerror(error, errbuf.data(), errbuf.size());
+
+    return -1;
+  }
+
+  return 0;
+}
+} // namespace
+
 int worker_process_event_loop(WorkerProcessConfig *wpconf) {
-  if (reopen_log_files() != 0) {
+  int rv;
+  std::array<char, STRERROR_BUFSIZE> errbuf;
+  (void)errbuf;
+
+  auto config = get_config();
+
+  if (reopen_log_files(config->logging) != 0) {
     LOG(FATAL) << "Failed to open log file";
     return -1;
   }
 
+  rv = ares_library_init(ARES_LIB_INIT_ALL);
+  if (rv != 0) {
+    LOG(FATAL) << "ares_library_init failed: " << ares_strerror(rv);
+    return -1;
+  }
+
   auto loop = EV_DEFAULT;
 
-  ConnectionHandler conn_handler(loop);
-
-  if (wpconf->server_fd6 != -1) {
-    conn_handler.set_acceptor6(
-        make_unique<AcceptHandler>(wpconf->server_fd6, &conn_handler));
-  }
-  if (wpconf->server_fd != 1) {
-    conn_handler.set_acceptor(
-        make_unique<AcceptHandler>(wpconf->server_fd, &conn_handler));
-  }
+  auto gen = util::make_mt19937();
 
 #ifdef HAVE_NEVERBLEED
-  if (!get_config()->upstream_no_tls || ssl::downstream_tls_enabled()) {
-    std::array<char, NEVERBLEED_ERRBUF_SIZE> errbuf;
-    auto nb = make_unique<neverbleed_t>();
-    if (neverbleed_init(nb.get(), errbuf.data()) != 0) {
-      LOG(FATAL) << "neverbleed_init failed: " << errbuf.data();
-      return -1;
-    }
-
-    LOG(NOTICE) << "neverbleed process [" << nb->daemon_pid << "] spawned";
-
-    conn_handler.set_neverbleed(std::move(nb));
+  std::array<char, NEVERBLEED_ERRBUF_SIZE> nb_errbuf;
+  auto nb = std::make_unique<neverbleed_t>();
+  if (neverbleed_init(nb.get(), nb_errbuf.data()) != 0) {
+    LOG(FATAL) << "neverbleed_init failed: " << nb_errbuf.data();
+    return -1;
   }
 
-  auto nb = conn_handler.get_neverbleed();
+  LOG(NOTICE) << "neverbleed process [" << nb->daemon_pid << "] spawned";
 
   ev_child nb_childev;
-  if (nb) {
-    ev_child_init(&nb_childev, nb_child_cb, nb->daemon_pid, 0);
-    nb_childev.data = nullptr;
-    ev_child_start(loop, &nb_childev);
-  }
 
+  ev_child_init(&nb_childev, nb_child_cb, nb->daemon_pid, 0);
+  nb_childev.data = nullptr;
+  ev_child_start(loop, &nb_childev);
 #endif // HAVE_NEVERBLEED
 
+  auto conn_handler = std::make_unique<ConnectionHandler>(loop, gen);
+
+#ifdef HAVE_NEVERBLEED
+  conn_handler->set_neverbleed(nb.get());
+#endif // HAVE_NEVERBLEED
+
+#ifdef ENABLE_HTTP3
+  conn_handler->set_quic_ipc_fd(wpconf->quic_ipc_fd);
+  conn_handler->set_quic_lingering_worker_processes(
+      wpconf->quic_lingering_worker_processes);
+#endif // ENABLE_HTTP3
+
+  for (auto &addr : config->conn.listener.addrs) {
+    conn_handler->add_acceptor(
+        std::make_unique<AcceptHandler>(&addr, conn_handler.get()));
+  }
+
+  MemchunkPool mcpool;
+
   ev_timer renew_ticket_key_timer;
-  if (!get_config()->upstream_no_tls) {
-    if (get_config()->tls_ticket_key_memcached_host) {
-      conn_handler.set_tls_ticket_key_memcached_dispatcher(
-          make_unique<MemcachedDispatcher>(
-              &get_config()->tls_ticket_key_memcached_addr, loop));
+  if (tls::upstream_tls_enabled(config->conn)) {
+    auto &ticketconf = config->tls.ticket;
+    auto &memcachedconf = ticketconf.memcached;
+
+    if (!memcachedconf.host.empty()) {
+      SSL_CTX *ssl_ctx = nullptr;
+
+      if (memcachedconf.tls) {
+        ssl_ctx = conn_handler->create_tls_ticket_key_memcached_ssl_ctx();
+      }
+
+      conn_handler->set_tls_ticket_key_memcached_dispatcher(
+          std::make_unique<MemcachedDispatcher>(
+              &ticketconf.memcached.addr, loop, ssl_ctx,
+              StringRef{memcachedconf.host}, &mcpool, gen));
 
       ev_timer_init(&renew_ticket_key_timer, memcached_get_ticket_key_cb, 0.,
                     0.);
-      renew_ticket_key_timer.data = &conn_handler;
+      renew_ticket_key_timer.data = conn_handler.get();
       // Get first ticket keys.
       memcached_get_ticket_key_cb(loop, &renew_ticket_key_timer, 0);
     } else {
       bool auto_tls_ticket_key = true;
-      if (!get_config()->tls_ticket_key_files.empty()) {
-        if (!get_config()->tls_ticket_key_cipher_given) {
+      if (!ticketconf.files.empty()) {
+        if (!ticketconf.cipher_given) {
           LOG(WARN)
               << "It is strongly recommended to specify "
                  "--tls-ticket-key-cipher=aes-128-cbc (or "
@@ -448,19 +521,18 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
                  "becomes aes-256-cbc";
         }
         auto ticket_keys = read_tls_ticket_key_file(
-            get_config()->tls_ticket_key_files,
-            get_config()->tls_ticket_key_cipher, EVP_sha256());
+            ticketconf.files, ticketconf.cipher, EVP_sha256());
         if (!ticket_keys) {
           LOG(WARN) << "Use internal session ticket key generator";
         } else {
-          conn_handler.set_ticket_keys(std::move(ticket_keys));
+          conn_handler->set_ticket_keys(std::move(ticket_keys));
           auto_tls_ticket_key = false;
         }
       }
       if (auto_tls_ticket_key) {
         // Generate new ticket key every 1hr.
         ev_timer_init(&renew_ticket_key_timer, renew_ticket_key_cb, 0., 1_h);
-        renew_ticket_key_timer.data = &conn_handler;
+        renew_ticket_key_timer.data = conn_handler.get();
         ev_timer_again(loop, &renew_ticket_key_timer);
 
         // Generate first session ticket key before running workers.
@@ -469,10 +541,69 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
     }
   }
 
-  int rv;
+#ifdef ENABLE_HTTP3
+  auto &quicconf = config->quic;
 
-  if (get_config()->num_worker == 1) {
-    rv = conn_handler.create_single_worker();
+  std::shared_ptr<QUICKeyingMaterials> qkms;
+
+  if (!quicconf.upstream.secret_file.empty()) {
+    qkms = read_quic_secret_file(quicconf.upstream.secret_file);
+    if (!qkms) {
+      LOG(WARN) << "Use QUIC keying materials generated internally";
+    }
+  }
+
+  if (!qkms) {
+    qkms = std::make_shared<QUICKeyingMaterials>();
+    qkms->keying_materials.resize(1);
+
+    auto &qkm = qkms->keying_materials.front();
+
+    if (RAND_bytes(qkm.reserved.data(), qkm.reserved.size()) != 1) {
+      LOG(ERROR) << "Failed to generate QUIC secret reserved data";
+      return -1;
+    }
+
+    if (RAND_bytes(qkm.secret.data(), qkm.secret.size()) != 1) {
+      LOG(ERROR) << "Failed to generate QUIC secret";
+      return -1;
+    }
+
+    if (RAND_bytes(qkm.salt.data(), qkm.salt.size()) != 1) {
+      LOG(ERROR) << "Failed to generate QUIC salt";
+      return -1;
+    }
+  }
+
+  for (auto &qkm : qkms->keying_materials) {
+    if (generate_quic_connection_id_encryption_key(
+            qkm.cid_encryption_key.data(), qkm.cid_encryption_key.size(),
+            qkm.secret.data(), qkm.secret.size(), qkm.salt.data(),
+            qkm.salt.size()) != 0) {
+      LOG(ERROR) << "Failed to generate QUIC Connection ID encryption key";
+      return -1;
+    }
+
+    qkm.cid_encryption_ctx = EVP_CIPHER_CTX_new();
+    if (!EVP_EncryptInit_ex(qkm.cid_encryption_ctx, EVP_aes_128_ecb(), nullptr,
+                            qkm.cid_encryption_key.data(), nullptr)) {
+      LOG(ERROR)
+          << "Failed to initialize QUIC Connection ID encryption context";
+      return -1;
+    }
+
+    EVP_CIPHER_CTX_set_padding(qkm.cid_encryption_ctx, 0);
+  }
+
+  conn_handler->set_quic_keying_materials(std::move(qkms));
+
+  conn_handler->set_cid_prefixes(wpconf->cid_prefixes);
+  conn_handler->set_quic_lingering_worker_processes(
+      wpconf->quic_lingering_worker_processes);
+#endif // ENABLE_HTTP3
+
+  if (config->single_thread) {
+    rv = conn_handler->create_single_worker();
     if (rv != 0) {
       return -1;
     }
@@ -484,12 +615,13 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
 
     rv = pthread_sigmask(SIG_BLOCK, &set, nullptr);
     if (rv != 0) {
-      LOG(ERROR) << "Blocking SIGCHLD failed: " << strerror(rv);
+      LOG(ERROR) << "Blocking SIGCHLD failed: "
+                 << xsi_strerror(rv, errbuf.data(), errbuf.size());
       return -1;
     }
 #endif // !NOTHREADS
 
-    rv = conn_handler.create_worker_thread(get_config()->num_worker);
+    rv = conn_handler->create_worker_thread(config->num_worker);
     if (rv != 0) {
       return -1;
     }
@@ -497,41 +629,82 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
 #ifndef NOTHREADS
     rv = pthread_sigmask(SIG_UNBLOCK, &set, nullptr);
     if (rv != 0) {
-      LOG(ERROR) << "Unblocking SIGCHLD failed: " << strerror(rv);
+      LOG(ERROR) << "Unblocking SIGCHLD failed: "
+                 << xsi_strerror(rv, errbuf.data(), errbuf.size());
       return -1;
     }
 #endif // !NOTHREADS
   }
 
+#if defined(ENABLE_HTTP3) && defined(HAVE_LIBBPF)
+  conn_handler->unload_bpf_objects();
+#endif // defined(ENABLE_HTTP3) && defined(HAVE_LIBBPF)
+
   drop_privileges(
 #ifdef HAVE_NEVERBLEED
-      nb
+      nb.get()
 #endif // HAVE_NEVERBLEED
-      );
+  );
 
   ev_io ipcev;
   ev_io_init(&ipcev, ipc_readcb, wpconf->ipc_fd, EV_READ);
-  ipcev.data = &conn_handler;
+  ipcev.data = conn_handler.get();
   ev_io_start(loop, &ipcev);
 
-  if (!get_config()->upstream_no_tls && !get_config()->no_ocsp) {
-    conn_handler.proceed_next_cert_ocsp();
+#ifdef ENABLE_HTTP3
+  ev_io quic_ipcev;
+  ev_io_init(&quic_ipcev, quic_ipc_readcb, wpconf->quic_ipc_fd, EV_READ);
+  quic_ipcev.data = conn_handler.get();
+  ev_io_start(loop, &quic_ipcev);
+#endif // ENABLE_HTTP3
+
+  if (tls::upstream_tls_enabled(config->conn) && !config->tls.ocsp.disabled) {
+    if (config->tls.ocsp.startup) {
+      conn_handler->set_enable_acceptor_on_ocsp_completion(true);
+      conn_handler->disable_acceptor();
+    }
+
+    conn_handler->proceed_next_cert_ocsp();
   }
 
   if (LOG_ENABLED(INFO)) {
     LOG(INFO) << "Entering event loop";
   }
 
+  if (send_ready_event(wpconf->ready_ipc_fd) != 0) {
+    return -1;
+  }
+
   ev_run(loop, 0);
 
-  conn_handler.cancel_ocsp_update();
+  conn_handler->cancel_ocsp_update();
+
+  // Destroy SSL_CTX held in conn_handler before killing neverbleed
+  // daemon.  Otherwise priv_rsa_finish yields "write error" and
+  // worker process aborts.
+  conn_handler.reset();
 
 #ifdef HAVE_NEVERBLEED
-  if (nb && nb->daemon_pid != -1) {
-    kill(nb->daemon_pid, SIGTERM);
+  assert(nb->daemon_pid > 0);
+
+  rv = kill(nb->daemon_pid, SIGTERM);
+  if (rv != 0) {
+    auto error = errno;
+    LOG(ERROR) << "Could not send signal to neverbleed daemon: errno=" << error;
+  }
+
+  while ((rv = waitpid(nb->daemon_pid, nullptr, 0)) == -1 && errno == EINTR)
+    ;
+  if (rv == -1) {
+    auto error = errno;
+    LOG(ERROR) << "Error occurred while we were waiting for the completion "
+                  "of neverbleed process: errno="
+               << error;
   }
 #endif // HAVE_NEVERBLEED
 
+  ares_library_cleanup();
+
   return 0;
 }
 
